// TODO main loop logic

// TODO BLE interface - incoming messages
// TODO RFID interface - read tag

#include <SPI.h>
#include "MFRC522.h"
#include <Wire.h>
#include "ACROBOTIC_SSD1306.h"
#include "logo.h"
#include "messages.h"
#include <SoftwareSerial.h>

// RFID module

#define RST_PIN         3          
#define SS_PIN          10         
MFRC522 mfrc522(SS_PIN, RST_PIN);  // Create MFRC522 instance
#define CARDIDSIZE      4          // MIFARE Classic tags have a 4 bytes UID
byte currentCardId[CARDIDSIZE];
byte previousCardId[CARDIDSIZE];
int cardStatus = 0;                // 0 = no card , 1 = card read, 2 = verifying, 3 = unknown user, 4 = credit low, 5 = access denied, 6 = access granted
char cardHolder[16];

// BLE module

SoftwareSerial mySerial(6, 7); // RX, TX  this is the UART link to the SKB369 module
#define BFSIZE 255
char incomingBuffer[255];
int ib_index = 0;
char incomingByte;
// char outgoingBuffer[BFSIZE];
int ob_index = 0;
int channelState = 0;               // 0 = not connected or unknown, 1 = advertising,  2 = pairing ,  3 = ready
char deviceMAC[20];
char deviceName[13] = "FB";
char gatewayMAC[] = "00:00:00:00:00"; 

// Timer mechanism for the heartbeat signal

unsigned long startMillis;  
unsigned long currentMillis;
const unsigned long period = 10000;  // in milliseconds

// SSR module
#define SSR_PIN         4  

// LEDs
#define RED_LED_PIN         A2 
#define YELLOW_LED_PIN      A1
#define GREEN_LED_PIN       A0 



// This code runs once, at startup

void setup() {
  Serial.begin(9600);
  Serial.println(F("Begin of setup")); 
  delay(3000);                           // to avoid sporadic I2C lockup, this is a known issue
  Wire.begin();                          // To connect to OLED over I2C
  oled.init();
  oled.drawBitmap(fablablogo, 1024);     // display logo
  setupRC522Module();                    // Initialize card reader
  setupSKB369Module();                   // Initialize BLE module  
  setupLEDs();
  delay(1000);
  startMillis = millis();                //initial start time
  Serial.println(F("End of setup"));
}

// This code runs continuously, after startup

void loop() 
{
  // Open communication channel with gateway
  openChannel();
  // Look whether any new and readable card has been swipped
  if (mfrc522.PICC_IsNewCardPresent() && mfrc522.PICC_ReadCardSerial()) 
  {
    readCard();
    sendCardCheckRequest(); 
  }
  // Update the SSR relay depending on the user access
  refreshSSR();
  // Update info on display and LEDs
  refreshDisplay();
  refreshLEDs();
  // Keep alive signal
  keepAlive();
}

// Update the SSR relay state depending on the access granted (or not .. ) to the user

void refreshSSR(){
  if (cardStatus == 6){
    digitalWrite(SSR_PIN, HIGH);
  }
  else{
    digitalWrite(SSR_PIN, LOW);
  }
}

// Clear the incoming data buffer

void clearIncomingBuffer(){
   ib_index = 0;
   for( int i = 0; i < 255;  ++i )
         incomingBuffer[i] = (char)0;
}

// Disconnect the device from the gateway

void disconnectDeviceFromGateway(){
  Serial.println(F("Disconnecting device from gateway..."));
  sendCommandToModule("AT+DISCON\r\n");
}

// On a regular basis, send a heartbeat signal to the gateway

void keepAlive(){
  currentMillis = millis();                   //get the current "time" (actually the number of milliseconds since the program started)
  if (currentMillis - startMillis >= period)  //test whether enough time has elapsed and we have to send the heartbeat signal 
  {
    // Serial.println(F("I m alive !"));
    sendMessage("HEARTBEAT");
    startMillis = currentMillis;   
  }
}

// Refresh information displayed on OLED screen

void refreshDisplay(){
  char line[16];
  oled.clearDisplay(); 
  oled.setFont(font8x8);
  // Line 1 =  Fab Buddy ID  
  oled.setTextXY(0,0); 
  oled.putString(deviceName);
  // Line 2 = intentionnaly empty       
  oled.setTextXY(1,0); 
  oled.putString("");
  // Line 3 = cardholder name
  oled.setTextXY(2,0); 
  oled.putString(cardHolder);
  // Line 4 = credit remaining
  oled.setTextXY(3,0); 
  oled.putString("");
  // Line 5 = RFID card status
  oled.setTextXY(4,0); 
  oled.putString(strcpy_P(line, (char*)pgm_read_word(&(messages[cardStatus + 24]))));
  // Line 6 = intentionnaly empty
  oled.setTextXY(5,0); 
  oled.putString("");
  // Line 7 = status of the SSR relay
  oled.setTextXY(6,0); 
  if (digitalRead(SSR_PIN) == HIGH){
    oled.putString(strcpy_P(line, (char*)pgm_read_word(&(messages[11]))));
  }
  else {
    oled.putString(strcpy_P(line, (char*)pgm_read_word(&(messages[12]))));
  }  
  // Line 8 = status of the gateway connection
  oled.setTextXY(7,0); 
  oled.putString(strcpy_P(line, (char*)pgm_read_word(&(messages[channelState]))));
  delay(1000);
}


// Incoming data may content a lot of different things. This routine helps to figure out 
// what kind of data it is and what to do with it

void parseIncomingBuffer(){
  //Serial.println("Parsing incoming buffer...");
  char msg[255];
  tranformIncomingBuffer(msg, sizeof(msg));
  //Serial.print("Msg = ");
  //Serial.println(msg);

  // Case 1 : a message indicating a host has disconnected
  
  if (strstr(msg, "Disconnected"))
  {
     channelState = 0;
     Serial.println(F("Disconnected from gateway"));
     clearIncomingBuffer();
     return;
  }

  // Case 2 : a message indicating a host is responding to our advertisment and trying to connect

  
  if (strstr(msg, "Connected"))
  {
     Serial.println(F("Connected to gateway"));
     clearIncomingBuffer();
     if (peerWithGateway()) channelState = 2;
     return;
  }

  // Case 3 : we get a message containing a MAC address. This should be interpreted as a peering completion 
  
  int cnt = 0;  
  for (int i=0; i<sizeof(msg); i++)  // once peering is done, the SKB send a [DA]zz:zz:zz:zz:zz:zz  where the variable part is the gateway MAC
  {
     if (msg[i] == ':') cnt++;
     if (cnt == 5){
        // TODO : add code to check the MAC address. If it doesn't fit the one we know the gateway has we should disconnect
        channelState = 3;
        Serial.println(F("Peering completed with gateway"));
        clearIncomingBuffer();
        return;
     }     
  }  

  // Case 4 : response to CARDCHECK   (3 = unknown user, 4 = credit low, 5 = access denied, 6 = access granted)

  if (strstr(msg, "CARDCHECK")){
     Serial.println(F("Response to CARDCHECK received"));
     if (strstr(msg, "ACCESSGRANTED")) cardStatus = 6;
     if (strstr(msg, "ACCESSDENIED")) cardStatus = 5;
     if (strstr(msg, "CREDITLOW")) cardStatus = 4;
     if (strstr(msg, "UNKNOWN")) cardStatus = 3;
     clearIncomingBuffer();
     return;
  }
  
  // Serial.println(msg);
}

// Peer with the gateway

bool peerWithGateway(){
  Serial.print(F("Peering request "));
  return sendCommandToModule("AT+PEER\r\n");
}

// Open the communication channel with the gateway

void openChannel(){  
  // To be polite, we listen first to any data sent from the network
  if (waitForData(1000)){
      parseIncomingBuffer();
  }
  // Now we talk
  switch (channelState)
  {
     case 0:  // idle. The SKB369 is on but does nothing actually. Let's start advertising ourselves on the network
        startAdvertising();
        break;
     case 1: // we are advertising, nothing else to do but to wait for the gateway to connect
        //Serial.println(F("Module is currently advertising, no message can be published..."));
        break; 
     case 2:  // gateway is connected, pairing request has been sent and we are waiting for an answer back to complete peering
        break;
     case 3:  // we are connected and peered to the gateway. Now we can (try to) talk...
        break;
  }     
}

// Send a message to the Gateway over the BLE link

bool sendMessage(char *message){
  if (channelState == 3){
    mySerial.write(message);
    return true;
  }
  return false;
}


// read RFID card identifier

void readCard(){
  //Serial.println(F("Card read"));
  for (int i = 0; i < CARDIDSIZE;  ++i ) {
     previousCardId[i] = currentCardId[i];
     currentCardId[i] = 0;
  }
  for (byte i = 0; i < mfrc522.uid.size; i++) {
    if (i < CARDIDSIZE) currentCardId[i] = mfrc522.uid.uidByte[i];
    cardStatus = 1;
    //Serial.print(currentCardId[i], HEX);
  }
}

// Refresh the current state of the communication channel

void refreshChannelState(){
  //Serial.println("Refreshing channel state... ");
  char connectionState[5];            // idle, adv(ertising), conn(ected) as per the BLE module
  sendCommandToModule("AT+WSTA\r\n");
  if (!waitForData(1000)) return;
  tranformIncomingBuffer(connectionState, sizeof(connectionState));  
  //Serial.print("Connection state = ");
  //Serial.println(connectionState);
  if (strstr(connectionState, "idle")){
     channelState = 0;
  }  
}

// Refresh the 3 LEDs on the front side

void refreshLEDs(){
  // Green LED
  if (channelState == 3){
    refreshLED(GREEN_LED_PIN, HIGH);
  }
  else
  {
    refreshLED(GREEN_LED_PIN, LOW);
  }
  // Yellow LED
  if (cardStatus == 4 ){
    refreshLED(YELLOW_LED_PIN, HIGH);
  }
  else
  {
    refreshLED(YELLOW_LED_PIN, LOW);
  }
  // Red LED
  if (cardStatus == 3 || cardStatus == 5 ){
    refreshLED(RED_LED_PIN, HIGH);
  }
  else
  {
    refreshLED(RED_LED_PIN, LOW);
  }
}

// This function set the right state for the rigth LED without unnecessary blink

void refreshLED( int pin, bool state){
  if (digitalRead(pin) != state) digitalWrite(pin, state);
}

// Send a request to the gateway to check a card Id

void sendCardCheckRequest(){
  Serial.print(F("Sending card check request: "));
  char buffer[32]; 
  for( int i = 0; i < 32;  ++i ) buffer[i] = (char)0;
  strcat(buffer, "CARDCHECK:");  
  for (unsigned int i = 0; i < CARDIDSIZE; i++){
     char buffer2[20];
     itoa(mfrc522.uid.uidByte[i], buffer2, 16);    
     // Serial.println(buffer2);    
     strcat(buffer, buffer2);
  }
  Serial.println(buffer);
  if (sendMessage(buffer)) cardStatus = 2; 
  // wait for an answer from the gateway
  if (waitForData(5000)) parseIncomingBuffer();
}

// Send command to the SBB369 module

bool sendCommandToModule(char *cmd){
  Serial.print(F("Command to send: ")); 
  Serial.println(cmd);
  clearIncomingBuffer();
  mySerial.write(cmd);
  delay(50);
  int timeout = 5000;
  while (timeout > 0){
     while (mySerial.available()){
        incomingByte = mySerial.read(); 
        // Serial.print(incomingByte);
        incomingBuffer[ib_index] = incomingByte;
        ib_index++;
        if (strstr(incomingBuffer, "[AT]OK") && strstr(incomingBuffer, "\n")) {
          //Serial.println("Accepted. ");
          clearIncomingBuffer();
          return true;
        } 
        if (strstr(incomingBuffer, "[AT]ER") && strstr(incomingBuffer, "\n")) {
          Serial.println(F("An error has occured. "));
          clearIncomingBuffer();
          return false;
        }  
     }     
     delay(50);
     timeout = timeout - 50;
  }  
  Serial.println(F("Command has timed out"));
  clearIncomingBuffer();
  return false;
}

// Setup the 3 LEDs on the front side and turn them OFF

void setupLEDs(){
  pinMode(RED_LED_PIN, OUTPUT);
  digitalWrite(RED_LED_PIN, LOW);
  pinMode(YELLOW_LED_PIN, OUTPUT);
  digitalWrite(YELLOW_LED_PIN, LOW);
  pinMode(GREEN_LED_PIN, OUTPUT);
  digitalWrite(GREEN_LED_PIN, LOW);
}

// Setup the RFID card reader

void setupRC522Module(){
  SPI.begin();                           // Init SPI bus
  mfrc522.PCD_Init();                    // Init MFRC522
  //if (mfrc522.PCD_PerformSelfTest()){
  //  Serial.println(F("Card reader is working"));
  //}
}

// Setup BLE module

void setupSKB369Module(){
  Serial.println(F("Pinging SKB369 module"));
  mySerial.begin(9600);  // set the data rate for the SoftwareSerial port, to communicate with SKB369 module
  if (sendCommandToModule("AT+VER\r\n")){  // just a simple health check
     waitForData(1000);                    // module will answer back something similar to [DA]B0169,V2.09,Jul 27 2018 15:52:34
     Serial.println(F("Communication established with SKB369 module"));
     setWirelessNetworkModuleID();         // set wireless module network ID to a well known name
  }
}

// Set a unique ID for this BLE device

void setWirelessNetworkModuleID(){
  Serial.println(F("Querying SKB369 module MAC address..."));
  sendCommandToModule("AT+MAC\r\n");
  if (waitForData(1000)){
     transformAddress(deviceMAC, sizeof(deviceMAC));   
     Serial.print(F("Current device MAC is "));
     Serial.println(deviceMAC);
     // Updating device name
     strcat(deviceName, deviceMAC);
     Serial.print(F("Updating device name, new name is "));
     Serial.println(deviceName);
     char cmd[] = "AT+DEV_NAME=";
     strcat(cmd, deviceName);
     strcat(cmd, "\r\n");
     sendCommandToModule(cmd);
     waitForData(1000);
  }
  else {
     Serial.print(F("Unable to get SKB369 module MAC address"));
  }
}

// Start advertising ourselves on the BLE network

void startAdvertising(){
  // Serial.print("Advertising ");
  if (sendCommandToModule("AT+ADV=start\r\n")) channelState = 1;
}

// Get rid of the [DA] prefix, the address separators and the CR-LF suffix

void transformAddress(char data[], int sizeofdata){
  //Serial.println("Transforming address..");
  int j = 0;
  for (int i=4; i<BFSIZE; i++){
     if (incomingBuffer[i] == '\n'){
          return;
     }
     if (incomingBuffer[i] != ':'){
         data[j] = incomingBuffer[i];
         j++; 
     }     
  } 
}

// Get rid of the [DA] prefix and the CR-LF suffix

void tranformIncomingBuffer(char data[], int sizeofdata){
     //Serial.println("Transforming buffer...");
     for (int i=4; i<BFSIZE; i++)
     {
        if (incomingBuffer[i] == char(0) || incomingBuffer[i] == '\n'){
          //Serial.println(data);
          //Serial.println(sizeofdata);
          return;
        }
        data[i-4] = incomingBuffer[i]; 
     }    
}

// Wait for data from the SKB369

bool waitForData(int timeOut){
  //Serial.println("Waiting for data");
  unsigned long stopAtMillis = millis() + timeOut;  
  clearIncomingBuffer();
  while (millis() <  stopAtMillis){
       while (mySerial.available()){
          incomingByte = mySerial.read(); 
          //Serial.print(incomingByte);
          incomingBuffer[ib_index] = incomingByte;
          ib_index++;
       }
       if (strstr(incomingBuffer, "[DA]") && strstr(incomingBuffer, "\n"))
       {
          Serial.print(F("Control Data received: "));
          Serial.println(incomingBuffer);
          return true;
       }  
       if (strstr(incomingBuffer, "}"))
       {
          Serial.print(F("Gateway Data received: "));
          Serial.println(incomingBuffer);
          return true;
       } 
       delay(50);
  }
  Serial.print("Time out expired after ");
  Serial.print(timeOut);
  Serial.println(" ms. No data received from SKB369");
  clearIncomingBuffer();
  return false;
}
